#   DrCr: Web-based double-entry bookkeeping framework
#   Copyright (C) 2022–2024  Lee Yingtong Li (RunasSudo)
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

from flask import abort, redirect, render_template, request, url_for

from .. import AMOUNT_DPS
from ..database import db
from ..models import Amount, Posting, Transaction
from ..webapp import app
from .models import StatementLine, StatementLineReconciliation

from datetime import datetime

@app.route('/statement-lines')
def statement_lines():
	# JOIN all associated postings (called in is_complex/charge_account)
	statement_lines = db.select(StatementLine).options(
		db.joinedload(StatementLine.reconciliation).joinedload(StatementLineReconciliation.posting).joinedload(Posting.transaction).joinedload(Transaction.postings)
	).order_by(StatementLine.dt.desc(), StatementLine.id.desc())
	
	if 'account' in request.args:
		statement_lines = statement_lines.where(StatementLine.source_account == request.args['account'])
	
	if request.args.get('unclassified', '0') == '1':
		statement_lines = statement_lines.where(StatementLine.reconciliation == None)
	
	page = db.paginate(statement_lines, per_page=int(request.args.get('per_page', 1000)))
	
	return render_template('statements/statement_lines.html', page=page)

@app.route('/statement-lines/charge', methods=['POST'])
def statement_line_charge():
	statement_line = db.session.get(StatementLine, request.form['line-id'])
	if not statement_line:
		abort(404)
	
	if statement_line.reconciliation:
		raise Exception('NYI')
	
	# Create transaction
	posting_source = Posting(account=statement_line.source_account, quantity=statement_line.quantity, commodity=statement_line.commodity)
	posting_charge = Posting(account=request.form['charge-account'], quantity=-statement_line.quantity, commodity=statement_line.commodity)
	transaction = Transaction(
		dt=statement_line.dt,
		description=statement_line.description,
		postings=[posting_source, posting_charge]
	)
	db.session.add(transaction)
	
	# Reconcile statement line
	reconciliation = StatementLineReconciliation(
		statement_line=statement_line,
		posting=posting_source
	)
	db.session.add(reconciliation)
	
	db.session.commit()
	
	return 'OK'

@app.route('/statement-lines/charge-complex', methods=['GET', 'POST'])
def statement_line_charge_complex():
	statement_line = db.session.get(StatementLine, request.args['line-id'])
	if not statement_line:
		abort(404)
	
	if statement_line.reconciliation:
		raise Exception('Cannot reconcile an already reconciled statement line')
	
	if request.method == 'GET':
		# Create template transaction
		posting_source = Posting(account=statement_line.source_account, quantity=statement_line.quantity, commodity=statement_line.commodity)
		posting_charge = Posting(account='', quantity=-statement_line.quantity, commodity=statement_line.commodity)
		transaction = Transaction(
			dt=statement_line.dt,
			description=statement_line.description,
			postings=[posting_source, posting_charge]
		)
		
		return render_template('journal/journal_edit_transaction.html', transaction=transaction)
	
	# New transaction
	transaction = Transaction(
		dt=datetime.strptime(request.form['dt'], '%Y-%m-%d'),
		description=request.form['description'],
		postings=[]
	)
	
	for account, sign, amount_str in zip(request.form.getlist('account'), request.form.getlist('sign'), request.form.getlist('amount')):
		amount = Amount.parse(amount_str)
		if sign == 'cr':
			amount = -amount
		
		posting = Posting(
			account=account,
			quantity=amount.quantity,
			commodity=amount.commodity
		)
		transaction.postings.append(posting)
	
	transaction.assert_valid()
	db.session.add(transaction)
	
	# Reconcile statement line
	posting_source = next((p for p in transaction.postings if p.account == statement_line.source_account and p.quantity == statement_line.quantity and p.commodity == statement_line.commodity), None)
	if not posting_source:
		raise Exception('Transaction must have posting corresponding to statement line')
	
	reconciliation = StatementLineReconciliation(
		statement_line=statement_line,
		posting=posting_source
	)
	db.session.add(reconciliation)
	
	db.session.commit()
	
	return redirect(url_for('statement_lines'))

@app.route('/statement-lines/reconcile-transfer', methods=['POST'])
def statement_line_reconcile_transfer():
	line_ids = request.form.getlist('sel-line-id')
	if len(line_ids) != 2:
		raise Exception('Must select exactly 2 statement lines')
	
	line1 = db.session.get(StatementLine, line_ids[0])
	line2 = db.session.get(StatementLine, line_ids[1])
	
	# Check same amount
	if line1.quantity != -line2.quantity or line1.commodity != line2.commodity:
		raise Exception('Selected statement line debits/credits must equal')
	
	if line1.reconciliation or line2.reconciliation:
		raise Exception('NYI')
	
	# Create transaction
	posting1 = Posting(account=line1.source_account, quantity=line1.quantity, commodity=line1.commodity)
	posting2 = Posting(account=line2.source_account, quantity=line2.quantity, commodity=line2.commodity)
	transaction = Transaction(
		dt=line1.dt,
		description=line1.description,
		postings=[posting1, posting2]
	)
	db.session.add(transaction)
	
	# Reconcile statement lines
	db.session.add(StatementLineReconciliation(
		statement_line=line1,
		posting=posting1
	))
	db.session.add(StatementLineReconciliation(
		statement_line=line2,
		posting=posting2
	))
	
	db.session.commit()
	
	return redirect(request.referrer or url_for('statement_lines'))

@app.route('/statement-lines/import', methods=['GET', 'POST'])
def statement_lines_import():
	if request.method == 'GET':
		return render_template('statements/import.html')
	
	# Import using importer
	if request.form['format'] == 'ofx1':
		from .importers.ofx1 import import_ofx1
		statement_lines = import_ofx1(request.files['file'])
	elif request.form['format'] == 'ofx2':
		from .importers.ofx2 import import_ofx2
		statement_lines = import_ofx2(request.files['file'])
	else:
		abort(400)
	
	# Fill in source_account
	for statement_line in statement_lines:
		statement_line.source_account = request.form['source-account']
	
	# Ignore pending, etc.
	# FIXME: This needs to be customisable
	statement_lines = [l for l in statement_lines if 'PENDING' not in l.description]
	
	if request.form['action'] == 'preview':
		return render_template('statements/import_preview.html', statement_lines=statement_lines)
	
	# Add to database
	for statement_line in statement_lines:
		db.session.add(statement_line)
	db.session.commit()
	
	return redirect(url_for('statement_lines'))
